#!/usr/bin/env nextflow

nextflow.enable.dsl=2

/* ChIPseq pipeline :
 * Pipeline that analyses ChIPseq data 
 * Preprocessing, filtration, alignment, peak calling.
*/

/* Params out */
params.fastp_out = "$params.project/fastp/"
params.index_fasta_out = "$params.project/Indexed_genome/"
params.sort_bam_out = "$params.project/Bam_filtered_sorted/"
params.index_bam_out = "$params.project/Bam_filt_sort_indexed/"
params.bam_to_bigwig_out = "$params.project/BigWig/" 
params.peak_calling_out = "$params.project/Peak_calling/"
params.bam_to_bed_out = "$params.project/Bed/"
params.bed_slop_out = "$params.project/Bed_sloped/"
params.bedGraph_out = "$params.project/BedGraph/"
params.chipseq_bam2BW_out = "$params.project/chipseq_BigGig"

/*
 ****************************************************************
                              Logs
 ****************************************************************
*/

log.info "fastq folder : ${params.fastq_folder}"
log.info "genome file : ${params.genome}"
log.info "genome sizes : ${params.chrom_sizes}"

/*
 ****************************************************************
                  Channel definitions
 ****************************************************************
*/

/* Raw paired-end reads fastq */
if (params.paired_end) {
  error "Not Implemented"
  /*
  Channel
    .fromFilePairs( params.fastq, size: 2 ) //def une error
    .set { fastq_files }
    */
} else {

  Channel
    .fromPath( params.input.collect { k, v -> "${params.fastq_folder}/${v.fastq}" })
    .ifEmpty { error "No fastq file defined" }
    .set { fastq_files }

  Channel
    .from( params.input.collect { k, v -> v.sample })
    .ifEmpty { error "No sample names given" }
    .set { sample_names }

  Channel
    .from( params.input.collect { k, v -> v.condition })
    .ifEmpty { error "No condition defined" }
    .set { condition_names }

  Channel
    .from( params.input.collect { k, v -> v.type })
    .ifEmpty { error "No sample types defined" }
    .set { sample_types }
}

Channel
  .fromPath( params.chrom_sizes )
  .ifEmpty { error "Cannot find any files matching: ${params.chrom_sizes}" }
  .map{it -> [(it.baseName =~ /([^\.]*)/)[0][1], it]}
  .set{ genome_sizes }

/*
 ****************************************************************
                          Imports
 ****************************************************************
*/

fastqc_mod = "./nf_modules/fastqc/main.nf"
include { fastp_chipster } from "./nf_modules/fastp/main.nf"
include { fastqc_fastq as fastqc_raw } from fastqc_mod addParams(fastqc_fastq_out: "$params.project/01_fastqc_raw/")
include { fastqc_fastq as fastqc_preprocessed } from fastqc_mod addParams(fastqc_fastq_out: "$params.project/02_fastqc_preprocessed/")
include { multiqc } from './nf_modules/multiqc/main.nf' addParams(multiqc_out: "$params.project/QC/")
include { index_fasta } from "./nf_modules/bowtie2/main.nf"
include { mapping_fastq_chipster } from "./nf_modules/bowtie2/main.nf"
include { filter_bam_quality_chipster } from "./nf_modules/samtools/main.nf"
include { sort_bam_chipster } from "./nf_modules/samtools/main.nf"
include { index_bam_chipster } from "./nf_modules/samtools/main.nf"
include { bam_to_bigwig } from "./nf_modules/deeptools/main.nf"
include { chipseq_bam2BW_chipster } from "./nf_modules/deeptools/main.nf"
include { peak_calling } from "./nf_modules/macs3/main.nf"

/*
 ****************************************************************
                          Workflow
 ****************************************************************
*/

workflow {

  //########################## PREPROCESSING ####################

  // fastp
  fastp_chipster(fastq_files, sample_names, condition_names, sample_types)
  
  //########################## QUALITY CHECKS ###################

  // fastqc_rawdata
  fastqc_raw(fastq_files.map { it -> [(it.baseName =~ /([^\.]*)/)[0][1], it] })
  // fastqc_processed
  fastqc_preprocessed(fastp_chipster.out.fastq.map { it -> [it [0], it[1]]})
  // multiqc
  multiqc(
    fastqc_raw.out.report
    .mix(
      fastqc_preprocessed.out.report
      ).collect()
  )

  //############ GENOME INDEXATION AND MAPPING ###################

  if (! params.idx) {
    Channel
      .fromPath( params.genome )
      .ifEmpty { error "Cannot find any files matching: ${params.genome}" }
      .map{it -> [(it.baseName =~ /([^\.]*)/)[0][1], it]}
      .set { genome_file }
    
    index_fasta(genome_file)
    mapping_fastq_chipster(index_fasta.out.index.collect(), fastp_chipster.out.fastq)
  }
  else {
    idx_genome = "${params.idx}*.bt2"
    Channel
      .fromPath( idx_genome )
      .ifEmpty { error "Cannot find idexed genome reference files" }
      .map { it -> [(it.baseName =~ /([^\.]*)/)[0][1], it ]}.groupTuple()
      .set { genome_indexed_input }
    mapping_fastq_chipster(genome_indexed_input.collect(), fastp_chipster.out.fastq)
  }
  
  //############### FILTERING, INDEXING, SORTING ##################

  // filter bam - remove reads with quality <30
  filter_bam_quality_chipster(mapping_fastq_chipster.out.bam)

  // samtools_sort
  sort_bam_chipster(filter_bam_quality_chipster.out.bam)

  // samtools_index
  index_bam_chipster(sort_bam_chipster.out.bam)

  //########################### BIGWIGS ###########################

  // Chipseq Bam 2 bigwig file with reads extends
  chipseq_bam2BW_chipster(index_bam_chipster.out.bam_idx)

  //######################## PEAK CALLING #########################

  // IP and Input samples identification and Channel creation
  index_bam_chipster.out.bam_idx.groupTuple(by: 3).set { combined_bams }
  combined_bams.map { it -> if(it[4][0] == 'IP') { [it[3], it[1][0], it[1][1]] } else {[ it[3], it[1][1], it[1][0]]} }.set { peak_calling_channel_in }

  // peak calling using MACS3 Prend des bed ou des bam en entrée...
  peak_calling(peak_calling_channel_in)
}